@@ -13,6 +13,7 @@ from __future__ import annotations
 import collections
 import contextlib
 import functools
+import importlib.machinery
 import io
 import itertools
 import json
@@ -31,23 +32,29 @@ import typing
 import warnings
 
 from typing import (
-    Any, ClassVar, DefaultDict, Dict, List, Optional, Set, TextIO, Tuple, Type,
-    Union
+    Any, ClassVar, DefaultDict, Dict, List, Optional, Sequence, Set, TextIO,
+    Tuple, Type, Union
 )
 
-import tomli
+
+if sys.version_info < (3, 11):
+    import tomli as tomllib
+else:
+    import tomllib
 
 import mesonpy._compat
 import mesonpy._elf
 import mesonpy._tags
 import mesonpy._util
+import mesonpy._wheelfile
 
-from mesonpy._compat import Collection, Iterator, Mapping, Path
+from mesonpy._compat import (
+    Collection, Iterator, Literal, Mapping, Path, typing_get_args
+)
 
 
 if typing.TYPE_CHECKING:  # pragma: no cover
     import pyproject_metadata  # noqa: F401
-    import wheel.wheelfile  # noqa: F401
 
 
 if sys.version_info >= (3, 8):
@@ -56,15 +63,7 @@ else:
     cached_property = lambda x: property(functools.lru_cache(maxsize=None)(x))  # noqa: E731
 
 
-__version__ = '0.10.0'
-
-
-class _depstr:
-    """Namespace that holds the requirement strings for dependencies we *might*
-    need at runtime. Having them in one place makes it easier to update.
-    """
-    patchelf = 'patchelf >= 0.11.0'
-    wheel = 'wheel >= 0.36.0'  # noqa: F811
+__version__ = '0.11.0'
 
 
 _COLORS = {
@@ -77,6 +76,15 @@ _COLORS = {
     'reset': '\33[0m',
 }
 _NO_COLORS = {color: '' for color in _COLORS}
+_NINJA_REQUIRED_VERSION = '1.8.2'
+
+
+class _depstr:
+    """Namespace that holds the requirement strings for dependencies we *might*
+    need at runtime. Having them in one place makes it easier to update.
+    """
+    patchelf = 'patchelf >= 0.11.0'
+    ninja = f'ninja >= {_NINJA_REQUIRED_VERSION}'
 
 
 def _init_colors() -> Dict[str, str]:
@@ -98,9 +106,9 @@ def _init_colors() -> Dict[str, str]:
 _STYLES = _init_colors()  # holds the color values, should be _COLORS or _NO_COLORS
 
 
-_LINUX_NATIVE_MODULE_REGEX = re.compile(r'^(?P<name>.+)\.(?P<tag>.+)\.so$')
-_WINDOWS_NATIVE_MODULE_REGEX = re.compile(r'^(?P<name>.+)\.(?P<tag>.+)\.pyd$')
-_STABLE_ABI_TAG_REGEX = re.compile(r'^abi(?P<abi_number>[0-9]+)$')
+_EXTENSION_SUFFIXES = frozenset(importlib.machinery.EXTENSION_SUFFIXES)
+_EXTENSION_SUFFIX_REGEX = re.compile(r'^\.(?:(?P<abi>[^.]+)\.)?(?:so|pyd|dll)$')
+assert all(re.match(_EXTENSION_SUFFIX_REGEX, x) for x in _EXTENSION_SUFFIXES)
 
 
 def _showwarning(
@@ -129,6 +137,10 @@ def _setup_cli() -> None:
         colorama.init()  # fix colors on windows
 
 
+class ConfigError(Exception):
+    """Error in the backend configuration."""
+
+
 class MesonBuilderError(Exception):
     """Error when building the Meson package."""
 
@@ -150,6 +162,7 @@ class _WheelBuilder():
     def __init__(
         self,
         project: Project,
+        metadata: Optional[pyproject_metadata.StandardMetadata],
         source_dir: pathlib.Path,
         install_dir: pathlib.Path,
         build_dir: pathlib.Path,
@@ -157,6 +170,7 @@ class _WheelBuilder():
         copy_files: Dict[str, str],
     ) -> None:
         self._project = project
+        self._metadata = metadata
         self._source_dir = source_dir
         self._install_dir = install_dir
         self._build_dir = build_dir
@@ -173,6 +187,11 @@ class _WheelBuilder():
     def _has_internal_libs(self) -> bool:
         return bool(self._wheel_files['mesonpy-libs'])
 
+    @property
+    def _has_extension_modules(self) -> bool:
+        # Assume that all code installed in {platlib} is Python ABI dependent.
+        return bool(self._wheel_files['platlib'])
+
     @property
     def basename(self) -> str:
         """Normalized wheel name and version (eg. meson_python-1.0.0)."""
@@ -181,14 +200,25 @@ class _WheelBuilder():
             version=self._project.version,
         )
 
+    @property
+    def tag(self) -> mesonpy._tags.Tag:
+        """Wheel tags."""
+        if self.is_pure:
+            return mesonpy._tags.Tag('py3', 'none', 'any')
+        if not self._has_extension_modules:
+            # The wheel has platform dependent code (is not pure) but
+            # does not contain any extension module (does not
+            # distribute any file in {platlib}) thus use generic
+            # implementation and ABI tags.
+            return mesonpy._tags.Tag('py3', 'none', None)
+        return mesonpy._tags.Tag(None, self._stable_abi, None)
+
     @property
     def name(self) -> str:
-        """Wheel name, this includes the basename and tags."""
-        return '{basename}-{python_tag}-{abi_tag}-{platform_tag}'.format(
+        """Wheel name, this includes the basename and tag."""
+        return '{basename}-{tag}'.format(
             basename=self.basename,
-            python_tag=self.python_tag,
-            abi_tag=self.abi_tag,
-            platform_tag=self.platform_tag,
+            tag=self.tag,
         )
 
     @property
@@ -220,12 +250,34 @@ class _WheelBuilder():
             Wheel-Version: 1.0
             Generator: meson
             Root-Is-Purelib: {is_purelib}
-            Tag: {tags}
+            Tag: {tag}
         ''').strip().format(
             is_purelib='true' if self.is_pure else 'false',
-            tags=f'{self.python_tag}-{self.abi_tag}-{self.platform_tag}',
+            tag=self.tag,
         ).encode()
 
+    @property
+    def entrypoints_txt(self) -> bytes:
+        """dist-info entry_points.txt."""
+        if not self._metadata:
+            return b''
+
+        data = self._metadata.entrypoints.copy()
+        data.update({
+            'console_scripts': self._metadata.scripts,
+            'gui_scripts': self._metadata.gui_scripts,
+        })
+
+        text = ''
+        for entrypoint in data:
+            if data[entrypoint]:
+                text += f'[{entrypoint}]\n'
+                for name, target in data[entrypoint].items():
+                    text += f'{name} = {target}\n'
+                text += '\n'
+
+        return text.encode()
+
     @property
     def _debian_python(self) -> bool:
         """Check if we are running on Debian-patched Python."""
@@ -239,166 +291,42 @@ class _WheelBuilder():
         except ModuleNotFoundError:
             return False
 
-    @property
-    def python_tag(self) -> str:
-        selected_tag = self._select_abi_tag()
-        if selected_tag and selected_tag.python:
-            return selected_tag.python
-        return 'py3'
-
-    @property
-    def abi_tag(self) -> str:
-        selected_tag = self._select_abi_tag()
-        if selected_tag:
-            return selected_tag.abi
-        return 'none'
-
     @cached_property
-    def platform_tag(self) -> str:
-        if self.is_pure:
-            return 'any'
-        # XXX: Choose the sysconfig platform here and let something like auditwheel
-        #      fix it later if there are system dependencies (eg. replace it with a manylinux tag)
-        platform_ = sysconfig.get_platform()
-        parts = platform_.split('-')
-        if parts[0] == 'macosx':
-            target = os.environ.get('MACOSX_DEPLOYMENT_TARGET')
-            if target:
-                print(
-                    '{yellow}MACOSX_DEPLOYMENT_TARGET is set so we are setting the '
-                    'platform tag to {target}{reset}'.format(target=target, **_STYLES)
-                )
-                parts[1] = target
-            else:
-                # If no target macOS version is specified fallback to
-                # platform.mac_ver() instead of sysconfig.get_platform() as the
-                # latter specifies the target macOS version Python was built
-                # against.
-                parts[1] = platform.mac_ver()[0]
-                if parts[1] >= '11':
-                    # Only pick up the major version, which changed from 10.X
-                    # to X.0 from macOS 11 onwards. See
-                    # https://github.com/FFY00/meson-python/issues/160
-                    parts[1] = parts[1].split('.')[0]
+    def _stable_abi(self) -> Optional[str]:
+        """Determine stabe ABI compatibility.
 
-            if parts[1] in ('11', '12'):
-                # Workaround for bug where pypa/packaging does not consider macOS
-                # tags without minor versions valid. Some Python flavors (Homebrew
-                # for example) on macOS started to do this in version 11, and
-                # pypa/packaging should handle things correctly from version 13 and
-                # forward, so we will add a 0 minor version to MacOS 11 and 12.
-                # https://github.com/FFY00/meson-python/issues/91
-                # https://github.com/pypa/packaging/issues/578
-                parts[1] += '.0'
+        Examine all files installed in {platlib} that look like
+        extension modules (extension .pyd on Windows, .dll on Cygwin,
+        and .so on other platforms) and, if they all share the same
+        PEP 3149 filename stable ABI tag, return it.
 
-            platform_ = '-'.join(parts)
-        elif parts[0] == 'linux' and parts[1] == 'x86_64' and sys.maxsize == 0x7fffffff:
-            # 32-bit Python running on an x86_64 host
-            # https://github.com/FFY00/meson-python/issues/123
-            parts[1] = 'i686'
-            platform_ = '-'.join(parts)
-        return platform_.replace('-', '_').replace('.', '_')
+        All files that look like extension modules are verified to
+        have a file name compatibel with what is expected by the
+        Python interpreter. An exception is raised otherwise.
 
-    def _calculate_file_abi_tag_heuristic_windows(self, filename: str) -> Optional[mesonpy._tags.Tag]:
-        """Try to calculate the Windows tag from the Python extension file name."""
-        match = _WINDOWS_NATIVE_MODULE_REGEX.match(filename)
-        if not match:
-            return None
-        tag = match.group('tag')
+        Other files are ignored.
 
-        try:
-            return mesonpy._tags.StableABITag(tag)
-        except ValueError:
-            return mesonpy._tags.InterpreterTag(tag)
-
-    def _calculate_file_abi_tag_heuristic_posix(self, filename: str) -> Optional[mesonpy._tags.Tag]:
-        """Try to calculate the Posix tag from the Python extension file name."""
-        # sysconfig is not guaranted to export SHLIB_SUFFIX but let's be
-        # preventive and check its value to make sure it matches our expectations
-        try:
-            extension = sysconfig.get_config_vars().get('SHLIB_SUFFIX', '.so')
-            if extension != '.so':
-                raise NotImplementedError(
-                    f"We don't currently support the {extension} extension. "
-                    'Please report this to https://github.com/FFY00/mesonpy/issues '
-                    'and include information about your operating system.'
-                )
-        except KeyError:
-            warnings.warn(
-                'sysconfig does not export SHLIB_SUFFIX, so we are unable to '
-                'perform the sanity check regarding the extension suffix. '
-                'Please report this to https://github.com/FFY00/mesonpy/issues '
-                'and include the output of `python -m sysconfig`.'
-            )
-        match = _LINUX_NATIVE_MODULE_REGEX.match(filename)
-        if not match:  # this file does not appear to be a native module
-            return None
-        tag = match.group('tag')
-
-        try:
-            return mesonpy._tags.StableABITag(tag)
-        except ValueError:
-            return mesonpy._tags.InterpreterTag(tag)
-
-    def _calculate_file_abi_tag_heuristic(self, filename: str) -> Optional[mesonpy._tags.Tag]:
-        """Try to calculate the ABI tag from the Python extension file name."""
-        if os.name == 'nt':
-            return self._calculate_file_abi_tag_heuristic_windows(filename)
-        # everything else *should* follow the POSIX way, at least to my knowledge
-        return self._calculate_file_abi_tag_heuristic_posix(filename)
-
-    def _file_list_repr(self, files: Collection[str], prefix: str = '\t\t', max_count: int = 3) -> str:
-        if len(files) > max_count:
-            files = list(itertools.islice(files, max_count)) + [f'(... +{len(files)}))']
-        return ''.join(f'{prefix}- {file}\n' for file in files)
-
-    def _files_by_tag(self) -> Mapping[mesonpy._tags.Tag, Collection[str]]:
-        """Map files into ABI tags."""
-        files_by_tag: Dict[mesonpy._tags.Tag, List[str]] = collections.defaultdict(list)
-
-        for _, file in self._wheel_files['platlib']:
-            # if in platlib, calculate the ABI tag
-            tag = self._calculate_file_abi_tag_heuristic(file)
-            if tag:
-                files_by_tag[tag].append(file)
-
-        return files_by_tag
-
-    def _select_abi_tag(self) -> Optional[mesonpy._tags.Tag]:  # noqa: C901
-        """Given a list of ABI tags, selects the most specific one.
-
-        Raises an error if there are incompatible tags.
         """
-        # Possibilities:
-        #   - interpreter specific (cpython/pypy/etc, version)
-        #   - stable abi (abiX)
-        tags = self._files_by_tag()
-        selected_tag = None
-        for tag, files in tags.items():
-            # no selected tag yet, let's assign this one
-            if not selected_tag:
-                selected_tag = tag
-            # interpreter tag
-            elif isinstance(tag, mesonpy._tags.InterpreterTag):
-                if tag != selected_tag:
-                    if isinstance(selected_tag, mesonpy._tags.InterpreterTag):
-                        raise ValueError(
-                            'Found files with incompatible ABI tags:\n'
-                            + self._file_list_repr(tags[selected_tag])
-                            + '\tand\n'
-                            + self._file_list_repr(files)
-                        )
-                    selected_tag = tag
-            # stable ABI
-            elif isinstance(tag, mesonpy._tags.StableABITag):
-                if isinstance(selected_tag, mesonpy._tags.StableABITag) and tag != selected_tag:
+        soext = sorted(_EXTENSION_SUFFIXES, key=len)[0]
+        abis = []
+
+        for path, src in self._wheel_files['platlib']:
+            if path.suffix == soext:
+                match = re.match(r'^[^.]+(.*)$', path.name)
+                assert match is not None
+                suffix = match.group(1)
+                if suffix not in _EXTENSION_SUFFIXES:
                     raise ValueError(
-                        'Found files with incompatible ABI tags:\n'
-                        + self._file_list_repr(tags[selected_tag])
-                        + '\tand\n'
-                        + self._file_list_repr(files)
-                    )
-        return selected_tag
+                        f'Extension module {str(path)!r} not compatible with Python interpreter. '
+                        f'Filename suffix {suffix!r} not in {set(_EXTENSION_SUFFIXES)}.')
+                match = _EXTENSION_SUFFIX_REGEX.match(suffix)
+                assert match is not None
+                abis.append(match.group('abi'))
+
+        stable = [x for x in abis if x and re.match(r'abi\d+', x)]
+        if len(stable) > 0 and len(stable) == len(abis) and all(x == stable[0] for x in stable[1:]):
+            return stable[0]
+        return None
 
     def _is_native(self, file: Union[str, pathlib.Path]) -> bool:
         """Check if file is a native file."""
@@ -528,7 +456,7 @@ class _WheelBuilder():
 
     def _install_path(
         self,
-        wheel_file: wheel.wheelfile.WheelFile,  # type: ignore[name-defined]
+        wheel_file: mesonpy._wheelfile.WheelFile,
         counter: mesonpy._util.CLICounter,
         origin: Path,
         destination: pathlib.Path,
@@ -571,18 +499,18 @@ class _WheelBuilder():
             wheel_file.write(origin, location)
 
     def build(self, directory: Path) -> pathlib.Path:
-        import wheel.wheelfile
-
         self._project.build()  # ensure project is built
 
         wheel_file = pathlib.Path(directory, f'{self.name}.whl')
 
-        with wheel.wheelfile.WheelFile(wheel_file, 'w') as whl:
+        with mesonpy._wheelfile.WheelFile(wheel_file, 'w') as whl:
             # add metadata
             whl.writestr(f'{self.distinfo_dir}/METADATA', self._project.metadata)
             whl.writestr(f'{self.distinfo_dir}/WHEEL', self.wheel)
+            if self.entrypoints_txt:
+                whl.writestr(f'{self.distinfo_dir}/entry_points.txt', self.entrypoints_txt)
 
-            # add license (see https://github.com/FFY00/meson-python/issues/88)
+            # add license (see https://github.com/mesonbuild/meson-python/issues/88)
             if self._project.license_file:
                 whl.write(
                     self._source_dir / self._project.license_file,
@@ -615,6 +543,10 @@ class _WheelBuilder():
         return wheel_file
 
 
+MesonArgsKeys = Literal['dist', 'setup', 'compile', 'install']
+MesonArgs = Mapping[MesonArgsKeys, Collection[str]]
+
+
 class Project():
     """Meson project wrapper to generate Python artifacts."""
 
@@ -628,15 +560,22 @@ class Project():
         source_dir: Path,
         working_dir: Path,
         build_dir: Optional[Path] = None,
+        meson_args: Optional[MesonArgs] = None,
     ) -> None:
         self._source_dir = pathlib.Path(source_dir).absolute()
         self._working_dir = pathlib.Path(working_dir).absolute()
         self._build_dir = pathlib.Path(build_dir).absolute() if build_dir else (self._working_dir / 'build')
         self._install_dir = self._working_dir / 'install'
         self._meson_native_file = self._source_dir / '.mesonpy-native-file.ini'
+        self._env = os.environ.copy()
+
+        # prepare environment
+        ninja_path = _env_ninja_command()
+        if ninja_path is not None:
+            self._env.setdefault('NINJA', str(ninja_path))
 
         # load config -- PEP 621 support is optional
-        self._config = tomli.loads(self._source_dir.joinpath('pyproject.toml').read_text())
+        self._config = tomllib.loads(self._source_dir.joinpath('pyproject.toml').read_text())
         self._pep621 = 'project' in self._config
         if self.pep621:
             try:
@@ -655,6 +594,17 @@ class Project():
         if self._metadata:
             self._validate_metadata()
 
+        # load meson args
+        self._meson_args = collections.defaultdict(tuple, meson_args or {})
+        for key in self._get_config_key('args'):
+            args_from_config = tuple(self._get_config_key(f'args.{key}'))
+            self._meson_args[key] = args_from_config + tuple(self._meson_args[key])
+        # XXX: We should validate the user args to make sure they don't conflict with ours.
+
+        self._check_for_unknown_config_keys({
+            'args': typing_get_args(MesonArgsKeys),
+        })
+
         # make sure the build dir exists
         self._build_dir.mkdir(exist_ok=True)
         self._install_dir.mkdir(exist_ok=True)
@@ -685,10 +635,21 @@ class Project():
         if self._metadata and 'version' in self._metadata.dynamic:
             self._metadata.version = self.version
 
+    def _get_config_key(self, key: str) -> Any:
+        value: Any = self._config
+        for part in f'tool.meson-python.{key}'.split('.'):
+            if not isinstance(value, Mapping):
+                raise ConfigError(
+                    f'Found unexpected value in `{part}` when looking for '
+                    f'config key `tool.meson-python.{key}` (`{value}`)'
+                )
+            value = value.get(part, {})
+        return value
+
     def _proc(self, *args: str) -> None:
         """Invoke a subprocess."""
         print('{cyan}{bold}+ {}{reset}'.format(' '.join(args), **_STYLES))
-        subprocess.check_call(list(args))
+        subprocess.check_call(list(args), env=self._env)
 
     def _meson(self, *args: str) -> None:
         """Invoke Meson."""
@@ -705,19 +666,18 @@ class Project():
             f'--prefix={sys.base_prefix}',
             os.fspath(self._source_dir),
             os.fspath(self._build_dir),
+            f'--native-file={os.fspath(self._meson_native_file)}',
+            # TODO: Allow configuring these arguments
+            '-Ddebug=false',
+            '-Doptimization=2',
+            # user args
+            *self._meson_args['setup'],
         ]
         if reconfigure:
             setup_args.insert(0, '--reconfigure')
 
         try:
-            self._meson(
-                'setup',
-                f'--native-file={os.fspath(self._meson_native_file)}',
-                # TODO: Allow configuring these arguments
-                '-Ddebug=false',
-                '-Doptimization=2',
-                *setup_args,
-            )
+            self._meson('setup', *setup_args)
         except subprocess.CalledProcessError:
             if reconfigure:  # if failed reconfiguring, try a normal configure
                 self._configure()
@@ -748,10 +708,22 @@ class Project():
                     f'expected `{self._metadata.requires_python}`'
                 )
 
+    def _check_for_unknown_config_keys(self, valid_args: Mapping[str, Collection[str]]) -> None:
+        config = self._config.get('tool', {}).get('meson-python', {})
+
+        for key, valid_subkeys in config.items():
+            if key not in valid_args:
+                raise ConfigError(f'Unknown configuration key: tool.meson-python.{key}')
+
+            for subkey in valid_args[key]:
+                if subkey not in valid_subkeys:
+                    raise ConfigError(f'Unknown configuration key: tool.meson-python.{key}.{subkey}')
+
     @cached_property
     def _wheel_builder(self) -> _WheelBuilder:
         return _WheelBuilder(
             self,
+            self._metadata,
             self._source_dir,
             self._install_dir,
             self._build_dir,
@@ -762,8 +734,8 @@ class Project():
     @functools.lru_cache(maxsize=None)
     def build(self) -> None:
         """Trigger the Meson build."""
-        self._meson('compile')
-        self._meson('install', '--destdir', os.fspath(self._install_dir))
+        self._meson('compile', *self._meson_args['compile'],)
+        self._meson('install', '--destdir', os.fspath(self._install_dir), *self._meson_args['install'],)
 
     @classmethod
     @contextlib.contextmanager
@@ -771,10 +743,11 @@ class Project():
         cls,
         source_dir: Path = os.path.curdir,
         build_dir: Optional[Path] = None,
+        meson_args: Optional[MesonArgs] = None,
     ) -> Iterator[Project]:
         """Creates a project instance pointing to a temporary working directory."""
         with tempfile.TemporaryDirectory(prefix='.mesonpy-', dir=os.fspath(source_dir)) as tmpdir:
-            yield cls(source_dir, tmpdir, build_dir)
+            yield cls(source_dir, tmpdir, build_dir, meson_args)
 
     @functools.lru_cache()
     def _info(self, name: str) -> Dict[str, Any]:
@@ -842,9 +815,8 @@ class Project():
         assert isinstance(version, str)
         return version
 
-    @property
-    @functools.lru_cache(maxsize=1)
-    def metadata(self) -> bytes:  # noqa: C901
+    @cached_property
+    def metadata(self) -> bytes:
         """Project metadata."""
         # the rest of the keys are only available when using PEP 621 metadata
         if not self.pep621:
@@ -882,7 +854,7 @@ class Project():
     def sdist(self, directory: Path) -> pathlib.Path:
         """Generates a sdist (source distribution) in the specified directory."""
         # generate meson dist file
-        self._meson('dist', '--allow-dirty', '--no-tests', '--formats', 'gztar')
+        self._meson('dist', '--allow-dirty', '--no-tests', '--formats', 'gztar', *self._meson_args['dist'],)
 
         # move meson dist file to output path
         dist_name = f'{self.name}-{self.version}'
@@ -903,6 +875,19 @@ class Project():
                     continue
                 path = self._source_dir.joinpath(*member_parts[1:])
 
+                if not path.exists() and member.isfile():
+                    # File doesn't exists on the source directory but exists on
+                    # the Meson dist, so it is generated file, which we need to
+                    # include.
+                    # See https://mesonbuild.com/Reference-manual_builtin_meson.html#mesonadd_dist_script
+
+                    # MESON_DIST_ROOT could have a different base name
+                    # than the actual sdist basename, so we need to rename here
+                    file = meson_dist.extractfile(member.name)
+                    member.name = str(pathlib.Path(dist_name, *member_parts[1:]).as_posix())
+                    tar.addfile(member, file)
+                    continue
+
                 if not path.is_file():
                     continue
 
@@ -925,8 +910,8 @@ class Project():
             pkginfo_info = tarfile.TarInfo(f'{dist_name}/PKG-INFO')
             if mtime:
                 pkginfo_info.mtime = mtime
-            pkginfo_info.size = len(self.metadata)  # type: ignore[arg-type]
-            tar.addfile(pkginfo_info, fileobj=io.BytesIO(self.metadata))  # type: ignore[arg-type]
+            pkginfo_info.size = len(self.metadata)
+            tar.addfile(pkginfo_info, fileobj=io.BytesIO(self.metadata))
 
         return sdist
 
@@ -945,12 +930,89 @@ def _project(config_settings: Optional[Dict[Any, Any]]) -> Iterator[Project]:
     if config_settings is None:
         config_settings = {}
 
+    # expand all string values to single element tuples and convert collections to tuple
+    config_settings = {
+        key: tuple(value) if isinstance(value, Collection) and not isinstance(value, str) else (value,)
+        for key, value in config_settings.items()
+    }
+
+    builddir_value = config_settings.get('builddir', {})
+    if len(builddir_value) > 0:
+        if len(builddir_value) != 1:
+            raise ConfigError('Specified multiple values for `builddir`, only one is allowed')
+        builddir = builddir_value[0]
+        if not isinstance(builddir, str):
+            raise ConfigError(f'Config option `builddir` should be a string (found `{type(builddir)}`)')
+    else:
+        builddir = None
+
+    def _validate_string_collection(key: str) -> None:
+        assert isinstance(config_settings, Mapping)
+        problematic_items: Sequence[Any] = list(filter(None, (
+            item if not isinstance(item, str) else None
+            for item in config_settings.get(key, ())
+        )))
+        if problematic_items:
+            raise ConfigError(
+                f'Config option `{key}` should only contain string items, but '
+                'contains the following parameters that do not meet this criteria:' +
+                ''.join((
+                    f'\t- {item} (type: {type(item)})'
+                    for item in problematic_items
+                ))
+            )
+
+    meson_args_keys = typing_get_args(MesonArgsKeys)
+    meson_args_cli_keys = tuple(f'{key}-args' for key in meson_args_keys)
+
+    for key in config_settings:
+        if key not in ('builddir', *meson_args_cli_keys):
+            raise ConfigError(f'Unknown config setting: {key}')
+
+    for key in meson_args_cli_keys:
+        _validate_string_collection(key)
+
     with Project.with_temp_working_dir(
-        build_dir=config_settings.get('builddir'),
+        build_dir=builddir,
+        meson_args=typing.cast(MesonArgs, {
+            key: config_settings.get(f'{key}-args', ())
+            for key in meson_args_keys
+        }),
     ) as project:
         yield project
 
 
+def _env_ninja_command(*, version: str = _NINJA_REQUIRED_VERSION) -> Optional[pathlib.Path]:
+    """
+    Returns the path to ninja, or None if no ninja found.
+    """
+    required_version = tuple(int(v) for v in version.split('.'))
+    env_ninja = os.environ.get('NINJA', None)
+    ninja_candidates = [env_ninja] if env_ninja else ['ninja', 'ninja-build', 'samu']
+    for ninja in ninja_candidates:
+        ninja_path = shutil.which(ninja)
+        if ninja_path is None:
+            continue
+
+        result = subprocess.run([ninja_path, '--version'], check=False, text=True, capture_output=True)
+
+        try:
+            candidate_version = tuple(int(x) for x in result.stdout.split('.')[:3])
+        except ValueError:
+            continue
+        if candidate_version < required_version:
+            continue
+        return pathlib.Path(ninja_path)
+
+    return None
+
+
+def get_requires_for_build_sdist(
+    config_settings: Optional[Dict[str, str]] = None,
+) -> List[str]:
+    return [_depstr.ninja] if _env_ninja_command() is None else []
+
+
 def build_sdist(
     sdist_directory: str,
     config_settings: Optional[Dict[Any, Any]] = None,
@@ -965,13 +1027,25 @@ def build_sdist(
 def get_requires_for_build_wheel(
     config_settings: Optional[Dict[str, str]] = None,
 ) -> List[str]:
-    dependencies = [_depstr.wheel]
-    with _project(config_settings) as project:
-        if not project.is_pure and platform.system() == 'Linux':
-            # we may need patchelf
-            if not shutil.which('patchelf'):  # XXX: This is slightly dangerous.
-                # patchelf not already acessible on the system
+    dependencies = []
+
+    if _env_ninja_command() is None:
+        dependencies.append(_depstr.ninja)
+
+    if sys.platform.startswith('linux'):
+        # we may need patchelf
+        if not shutil.which('patchelf'):
+            # patchelf not already accessible on the system
+            if _env_ninja_command() is not None:
+                # we have ninja available, so we can run Meson and check if the project needs patchelf
+                with _project(config_settings) as project:
+                    if not project.is_pure:
+                        dependencies.append(_depstr.patchelf)
+            else:
+                # we can't check if the project needs patchelf, so always add it
+                # XXX: wait for https://github.com/mesonbuild/meson/pull/10779
                 dependencies.append(_depstr.patchelf)
+
     return dependencies
 
 
